import sys

from PySide6 import QtWidgets

from castoranalytics.ui.mainwindow import MainWindow
from castoranalytics.ui.bootstrap import BOOTSTRAP_CSS


def main():
    QtWidgets.QApplication.setApplicationName('Castor Analytics')
    app = QtWidgets.QApplication(sys.argv)
    app.setStyleSheet(BOOTSTRAP_CSS)
    main_window = MainWindow()
    sys.exit(app.exec())


if __name__ == '__main__':
    main()


class CountryCodes:
    def __init__(self):
        self._codes = {}

    def add_country(self, country):
        self._codes[country.get_id()] = {
            'tld': country.get_tld_code(),
            'two_digit': country.get_two_digit_code(),
            'three_digit': country.get_three_digit_code(),
        }

    def get_tld_for(self, country_id):
        return self._codes[country_id]['tld']
    
    def get_two_digit_code_for(self, country_id):
        return self._codes[country_id]['two_digit']
    
    def get_three_digit_code_for(self, country_id):
        return self._codes[country_id]['three_digit']
    
    def __str__(self):
        return f'CountryCodes(codes={self._codes})'


from PySide6.QtWidgets import (
    QPushButton,
    QTableWidget,
    QTableWidgetItem,
    QHeaderView,
    QSizePolicy,
)
from PySide6.QtCore import Qt

import castoranalytics.ui.constants as constants

from castoranalytics.ui.pages.basepage import BasePage
from castoranalytics.ui.utils import Label
from castoranalytics.core.logging import LogManager

LOG = LogManager()


class StudyPage(BasePage):
    def __init__(self):
        super(StudyPage, self).__init__(name='Study')
        self._study_id = None
        self._back_button = self.init_back_button()
        self._study_name_label = self.init_study_name_label()
        self._table_widget = self.init_table_widget()
        self._sites_warning_label = self.init_sites_warning_label()
        self._show_sites_button = self.init_show_sites_button()
        self.init_page_layout()

    # INITIALIZATION

    def init_back_button(self):
        button = QPushButton('Back', self)
        button.clicked.connect(self.on_back)
        return button
    
    def init_table_widget(self):
        widget = QTableWidget()
        widget.setSelectionMode(QTableWidget.NoSelection)
        widget.horizontalHeader().setVisible(False)
        widget.verticalHeader().setVisible(False)
        widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        widget.setFocusPolicy(Qt.NoFocus)
        return widget
    
    def init_study_name_label(self):
        label = Label('', type=Label.HEADING1)
        return label
    
    def init_sites_warning_label(self):
        label = Label(constants.CASTOR_ANALYTICS_STUDY_SITES_WARNING, type=Label.HEADING1, style='color: red;')
        label.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        label.setContentsMargins(0, 0, 0, 20)
        label.setVisible(False)
        return label
    
    def init_show_sites_button(self):
        button = QPushButton('Get sites', self)
        button.clicked.connect(self.on_show_sites)
        button.setVisible(False)
        return button

    def init_page_layout(self):
        self.get_layout().addWidget(self._back_button)
        self.get_layout().addWidget(self._study_name_label)
        self.get_layout().addWidget(self._table_widget)
        self.get_layout().addWidget(self._sites_warning_label)
        self.get_layout().addWidget(self._show_sites_button)

    # TABLE

    def update_study_name_label(self, study, error):
        if error:
            self._study_name_label.setText(error)
            return
        self._study_name_label.setText(study.get_name())

    def update_table_widget(self, study):
        self._table_widget.setRowCount(5)
        self._table_widget.setColumnCount(2)
        self._table_widget.setItem(0, 0, QTableWidgetItem('Study ID'))
        self._table_widget.setItem(0, 1, QTableWidgetItem(study.get_id()))
        self._table_widget.setItem(1, 0, QTableWidgetItem('Number of sites'))
        self._table_widget.setItem(1, 1, QTableWidgetItem(str(study.get_nr_sites())))
        self._table_widget.setItem(2, 0, QTableWidgetItem('Number of participants'))
        self._table_widget.setItem(2, 1, QTableWidgetItem(str(study.get_nr_participants())))
        self._table_widget.setItem(3, 0, QTableWidgetItem('Number of fields'))
        self._table_widget.setItem(3, 1, QTableWidgetItem(str(study.get_nr_fields())))
        self._table_widget.setItem(4, 0, QTableWidgetItem('Created on'))
        self._table_widget.setItem(4, 1, QTableWidgetItem(str(study.get_created_on())))
        self._table_widget.setAlternatingRowColors(True)
        self._table_widget.horizontalHeader().setSectionResizeMode(0, QHeaderView.Stretch)
        self._table_widget.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self._table_widget.horizontalHeader().setDefaultAlignment(Qt.AlignLeft | Qt.AlignVCenter)

    def show_sites_button(self, study):
        if study.get_nr_sites() > 10:
            self._sites_warning_label.setVisible(True)
        self._show_sites_button.setVisible(True)
        
    # EVENT HANDLERS

    def on_back(self):
        self.navigate(f'/studies')

    def on_show_sites(self):
        self.navigate(f'/studies/{self._study_id}/sites')

    def on_navigate(self, params):
        self._study_name_label.setText('')
        self._table_widget.clearContents()
        self._sites_warning_label.setVisible(False)
        self._show_sites_button.setVisible(False)
        self._study_id = params.get('study_id', None)
        if self._study_id:
            self.load_data('get_study', self._study_id)

    def on_data_ready(self, study, error):
        self.update_study_name_label(study, error)
        self.update_table_widget(study)
        self.show_sites_button(study)


import io
import csv

from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session


class CastorApiClient:
    def __init__(self, client_id, client_secret, token_url, api_base_url):
        self._session = self.create_session(client_id, client_secret, token_url)
        self._api_url = api_base_url

    def create_session(self, client_id, client_secret, token_url):
        client = BackendApplicationClient(client_id=client_id)
        client_session = OAuth2Session(client=client)
        client_session.fetch_token(token_url=token_url, code=client_id, client_secret=client_secret)
        return client_session
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._session.close()

    # HELPERS

    def load_csv_data_as_dict(self, csv_data):
        dict_data = []
        for row in csv.DictReader(io.StringIO(csv_data), delimiter=';'):
            dict_data.append(row)
        return dict_data

    # COUNTRIES

    def get_countries(self):
        countries = []
        response = self._session.get(f'{self._api_url}/country')
        response.raise_for_status()
        response_data = response.json()
        for item in response_data.get('results', []):
            countries.append(item)
        # for item in countries:
        #     print(item)
        return countries
    
    # STUDIES
    
    def get_studies(self):
        studies = []
        response = self._session.get(f'{self._api_url}/study')
        response.raise_for_status()
        response_data = response.json()
        for study in response_data.get('_embedded', {}).get('study', {}):
            studies.append(study)
        # for item in studies:
        #     print(item)
        return studies
    
    def get_study(self, study_id):
        response = self._session.get(f'{self._api_url}/study/{study_id}')
        response.raise_for_status()
        response_data = response.json()
        return response_data
    
    # SITES
    
    def get_number_of_sites(self, study_id):
        response = self._session.get(f'{self._api_url}/study/{study_id}/site?page=1')
        response.raise_for_status()
        response_data = response.json()
        return int(response_data.get('total_items', 0))

    def get_study_sites(self, study_id):
        study_sites, current_page = [], 1
        while True:
            new_sites, page_count = self.get_study_sites_by_page(study_id, current_page)
            if not new_sites:
                break
            study_sites.extend(new_sites)
            if current_page >= page_count:
                break
            current_page += 1
        # for item in study_sites:
        #     print(item)
        return study_sites
    
    def get_study_sites_by_page(self, study_id, page):
        response = self._session.get(f'{self._api_url}/study/{study_id}/site?page={page}')
        response.raise_for_status()
        response_data = response.json()
        sites = response_data.get('_embedded', {}).get('sites', [])
        page_count = response_data.get('page_count', 1)
        return sites, page_count
    
    def get_study_site(self, study_id, site_id):
        response = self._session.get(f'{self._api_url}/study/{study_id}/site/{site_id}')
        response.raise_for_status()
        response_data = response.json()
        return response_data

    def get_statistics(self, study_id):
        response = self._session.get(f'{self._api_url}/study/{study_id}/statistics')
        response.raise_for_status()
        response_data = response.json()
        return response_data
    
    # PARTICIPANTS
    
    def get_participants(self, study_id):
        participants, current_page = [], 1
        while True:
            new_participants, page_count = self.get_participants_by_page(study_id, current_page)
            if not new_participants:
                break
            participants.extend(new_participants)
            if current_page >= page_count:
                break
            current_page += 1
        return participants
    
    def get_participants_by_page(self, study_id, page):
        response = self._session.get(f'{self._api_url}/study/{study_id}/participant?page={page}')
        response.raise_for_status()
        response_data = response.json()
        new_participants = response_data.get('_embedded', {}).get('participants', [])
        page_count = response_data.get('page_count', 1)
        return new_participants, page_count
    
    # PARTICIPANT PROGRESS

    def get_participant_progress(self, study_id):
        participants_progress_data, current_page = [], 1
        while True:
            new_participant_progress_data, page_count = self.get_participant_progress_by_page(study_id, current_page)
            if not new_participant_progress_data:
                break
            participants_progress_data.extend(new_participant_progress_data)
            if current_page >= page_count:
                break
            current_page += 1
        return participants_progress_data
    
    def get_participant_progress_by_page(self, study_id, page):
        response = self._session.get(f'{self._api_url}/study/{study_id}/participant-progress/forms?page={page}')
        response.raise_for_status()
        response_data = response.json()
        new_participant_progress_data = response_data.get('_embedded', {}).get('participants', [])
        page_count = response_data.get('page_count', 1)
        return new_participant_progress_data, page_count

    # DATA

    def get_fields(self, study_id):
        response = self._session.get(f'{self._api_url}/study/{study_id}/export/structure')
        response.raise_for_status()
        return self.load_csv_data_as_dict(response.text)

    def get_optiongroups(self, study_id):
        response = self._session.get(f'{self._api_url}/study/{study_id}/export/optiongroups')
        response.raise_for_status()
        return response.text

    def get_records(self, study_id):
        response = self._session.get(f'{self._api_url}/study/{study_id}/export/data')
        response.raise_for_status()
        return response.text



from PySide6.QtWidgets import QStackedWidget

from castoranalytics.ui.pages.route import Route


class Router(QStackedWidget):
    def __init__(self):
        super(Router, self).__init__()
        self._pages = {}
        self._routes = []
        self._prev_path = None
        self._curr_path = None

    def add_page(self, page, path_pattern):
        route = Route(page, path_pattern)
        self._routes.append(route)
        page.set_router(self)
        self.addWidget(page)

    def back(self):
        self.navigate(self._prev_path)

    def navigate(self, path):
        for route in self._routes:
            params = route.match(path)
            if params is not None:
                if path != self._curr_path:
                    self._prev_path = self._curr_path
                    self._curr_path = path
                    self.setCurrentWidget(route.get_page())
                    route.get_page().on_navigate(params)


from castoranalytics.core.api.study import Study


class StudyDetails(Study):
    def __init__(self, study_data):
        super(StudyDetails, self).__init__(study_data)

    def get_nr_sites(self):
        return self._study_data['nr_sites']

    def get_nr_participants(self):
        return self._study_data['nr_participants']

    def get_nr_fields(self):
        return self._study_data['nr_fields']
    
    def __str__(self):
        return 'StudyDetails(id={}, name={}, created_on={}, nr_sites={}, nr_participants={}, nr_fields={})'.format(
            self.get_id(),
            self.get_name(),
            self.get_created_on(),
            self.get_nr_sites(),
            self.get_nr_participants(),
            self.get_nr_fields(),
        )


BOOTSTRAP_CSS = """
QPushButton {
    background-color: #0d6efd;
    color: white;
    border: none;
    padding: 6px 12px;
    border-radius: 4px;
    font-size: 14px;
    font-weight: bold;
}

QPushButton:hover {
    background-color: #0b5ed7;
}

QPushButton:pressed {
    background-color: #0a58ca;
}

QTableWidget {
    font-size: 12px;
    font-weight: bold;
    color: slateblue;
}

QHeaderView::section {
    color: darkslateblue;
    font-size: 14px;
    font-weight: bold;
    padding: 4px;
}

"""


import os
import sys

from PySide6.QtWidgets import (
    QWidget, 
    QLabel, 
    QVBoxLayout, 
    QHBoxLayout, 
    QFrame,
)
from PySide6.QtCore import (
    Qt, 
    QEvent,
    QObject,
    QCoreApplication,
    QSettings,
)


class Settings(QSettings):
    def __init__(self):
        super(Settings, self).__init__(QSettings.IniFormat, QSettings.UserScope, 'com.rbeesoft', 'castoranalytics')

    def get(self, name, default=None):
        value = self.value(name)
        if value is None or value == '':
            return default
        return value

    def set(self, name, value):
        self.setValue(name, value)


class FunctionEvent(QEvent):
    EVENT_TYPE = QEvent.Type(QEvent.registerEventType())

    def __init__(self, func, *args, **kwargs):
        super().__init__(FunctionEvent.EVENT_TYPE)
        self.func = func
        self.args = args
        self.kwargs = kwargs


def is_macos():
    return sys.platform.startswith('darwin')


class EventDispatcher(QObject):
    def event(self, event):
        if isinstance(event, FunctionEvent):
            event.func(*event.args, **event.kwargs)
            return True
        return super().event(event)


_dispatcher = EventDispatcher()


def to_main_thread(func):
    def wrapper(*args, **kwargs):
        QCoreApplication.postEvent(_dispatcher, FunctionEvent(func, *args, **kwargs))
    return wrapper


def resource_path(relative_path):
    try:
        base_path = sys._MEIPASS
    except AttributeError:
        base_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
    return os.path.join(base_path, relative_path)


class BusyOverlay(QWidget):
    def __init__(self, parent=None, message="Loading..."):
        super().__init__(parent)
        self.setAttribute(Qt.WA_TransparentForMouseEvents, False)
        self.setAttribute(Qt.WA_NoSystemBackground)
        self.setAttribute(Qt.WA_StyledBackground)
        self.setWindowFlags(Qt.Widget | Qt.FramelessWindowHint)
        self.setStyleSheet("""
            background-color: rgba(0, 0, 0, 128);
        """)
        box = QFrame(self)
        box.setStyleSheet("""
            background-color: rgba(256, 256, 256, 200);
            color: white;
            border-radius: 10px;
            padding: 15px;
        """)
        box.setFrameShape(QFrame.StyledPanel)
        label = QLabel(message, box)
        label.setStyleSheet("font-size: 16px;")
        label.setAlignment(Qt.AlignCenter)
        box_layout = QVBoxLayout(box)
        box_layout.addWidget(label)
        main_layout = QHBoxLayout(self)
        main_layout.addStretch()
        vbox = QVBoxLayout()
        vbox.addStretch()
        vbox.addWidget(box, alignment=Qt.AlignCenter)
        vbox.addStretch()
        main_layout.addLayout(vbox)
        main_layout.addStretch()
        self.setLayout(main_layout)
        if parent:
            parent.installEventFilter(self)
        self.hide()

    def show_overlay(self):
        if self.parent():
            self.resize(self.parent().size())
        self.show()
        self.raise_()

    def hide_overlay(self):
        self.hide()

    def eventFilter(self, obj, event):
        if obj == self.parent() and event.type() == QEvent.Resize:
            self.resize(self.parent().size())
        return super().eventFilter(obj, event)
    

class Label(QLabel):
    HEADING1 = 1
    HEADING2 = 2
    HEADING3 = 3

    def __init__(self, text, type=None, style=''):
        super(Label, self).__init__(text)
        if type == self.HEADING1:
            style = f'font-size: 18px; font-weight: bold; color: darkslateblue; {style}'
        elif type == self.HEADING2:
            style = f'font-size: 16px; font-weight: bold; color: darkslateblue; {style}'
        elif type == self.HEADING3:
            style = f'font-size: 14px; font-weight: bold; color: darkslateblue; {style}'
        else:
            style = f'font-size: 12px; color: slateblue; {style}'
        self.setStyleSheet(style)


from PySide6.QtWidgets import QTableWidgetItem


class NumericTableWidgetItem(QTableWidgetItem):
    def __init__(self, value):
        super().__init__(str(value))
        self._value = value

    def __lt__(self, other):
        if isinstance(other, NumericTableWidgetItem):
            return self._value < other._value
        return super().__lt__(other)


from castoranalytics.core.singleton import singleton


@singleton
class Cache:
    def __init__(self):
        self._cache_items = {}

    def get(self, name, default=None):
        if name in self._cache_items.keys():
            return self._cache_items[name]
        return default
    
    def set(self, name, value):
        self._cache_items[name] = value

    def clear(self):
        del self._cache_items


import json
import threading

from concurrent.futures import ThreadPoolExecutor

from castoranalytics.core.singleton import singleton
from castoranalytics.core.api.study import Study
from castoranalytics.core.api.studydetails import StudyDetails
from castoranalytics.core.api.studysite import StudySite
from castoranalytics.core.api.country import Country
from castoranalytics.core.api.countrycodes import CountryCodes
from castoranalytics.core.logging import LogManager
from castoranalytics.core.cache import Cache
from castoranalytics.core.api.castorapiclient import CastorApiClient
from castoranalytics.core.utils import current_time_in_seconds, elapsed_time_in_seconds

LOG = LogManager()


@singleton
class Core:
    def __init__(self, client_id, client_secret, token_url, api_base_url):
        self._client_id = client_id
        self._client_secret = client_secret
        self._token_url = token_url
        self._api_base_url = api_base_url
        self._executor = ThreadPoolExecutor(max_workers=4)
        self._callbacks_lock = threading.Lock()
        self._callbacks = {}
        self._cache = Cache()

    # HELPERS

    def ready(self):
        return self._client_id is not None and self._client_secret is not None and self._token_url is not None and self._api_base_url is not None
    
    def set_api_credentials(self, client_id, client_secret, token_url, api_base_url):
        self._client_id = client_id
        self._client_secret = client_secret
        self._token_url = token_url
        self._api_base_url = api_base_url
    
    def read_from_cache(self, name):
        return self._cache.get(name)
    
    def write_to_cache(self, name, value):
        self._cache.set(name, value)

    # COUNTRIES
    
    def get_countries(self):
        start_time = current_time_in_seconds()
        countries = self.read_from_cache('countries')
        if countries:
            # LOG.info(f'Core.get_countries(): {elapsed_time_in_seconds(start_time)}')
            return countries
        with CastorApiClient(self._client_id, self._client_secret, self._token_url, self._api_base_url) as client:
            countries = []
            for country_data in client.get_countries():
                countries.append(Country(country_data))
            self.write_to_cache('countries', countries)
            LOG.info(f'Core.get_countries(): {elapsed_time_in_seconds(start_time)}')
            return countries
        
    def get_country_codes(self):
        start_time = current_time_in_seconds()
        country_codes = self.read_from_cache('country_codes')
        if country_codes:
            # LOG.info(f'Core.get_country_codes(): {elapsed_time_in_seconds(start_time)}')
            return country_codes
        countries = self.get_countries()
        country_codes = CountryCodes()
        for country in countries:
            country_codes.add_country(country)
        self.write_to_cache('country_codes', country_codes)
        LOG.info(f'Core.get_country_codes(): {elapsed_time_in_seconds(start_time)}')
        return country_codes
    
    # STUDIES

    def get_studies(self):
        start_time = current_time_in_seconds()
        studies = self.read_from_cache('studies')
        if studies:
            # LOG.info(f'Core.get_studies(): {elapsed_time_in_seconds(start_time)}')
            return studies
        with CastorApiClient(self._client_id, self._client_secret, self._token_url, self._api_base_url) as client:
            studies = []
            for study_data in client.get_studies():
                studies.append(Study(study_data))
            self.write_to_cache('studies', studies)
            LOG.info(f'Core.get_studies(): {elapsed_time_in_seconds(start_time)}')
            return studies
        
    def get_study(self, study_id):
        start_time = current_time_in_seconds()
        study = self.read_from_cache(f'studies/{study_id}')
        if study:
            # LOG.info(f'Core.get_study(): {elapsed_time_in_seconds(start_time)}')
            return study
        with CastorApiClient(self._client_id, self._client_secret, self._token_url, self._api_base_url) as client:
            study_data = client.get_study(study_id)
            study_statistics = client.get_statistics(study_id)
            fields = client.get_fields(study_id)
            study_data['nr_sites'] = client.get_number_of_sites(study_id)
            study_data['nr_participants'] = study_statistics['records']['total_count']
            study_data['nr_fields'] = len(fields)
            study = StudyDetails(study_data)
            self.write_to_cache(f'studies/{study_id}', study)
            LOG.info(f'Core.get_study(): {elapsed_time_in_seconds(start_time)}')
            return study
        
    # SITES

    def get_participant_site_abbreviations(self, participants):
        site_abbreviations = {}
        for participant in participants:
            site_abbreviations[participant['participant_id']] = participant['_embedded']['site']['abbreviation']
        return site_abbreviations
    
    def get_nr_records_per_site(self, participant_progress, participant_site_abbreviations):
        nr_records_per_site = {}
        for item in participant_progress:
            site_abbreviation = participant_site_abbreviations[item['participant_id']]
            if site_abbreviation not in nr_records_per_site.keys():
                nr_records_per_site[site_abbreviation] = 0
            nr_records_per_site[site_abbreviation] += 1
        return nr_records_per_site
    
    def get_completion_percentages_per_site(self, participant_progress, participant_site_abbreviations):
        completion_percentages_per_site = {}
        for item in participant_progress:
            site_abbreviation = participant_site_abbreviations[item['participant_id']]
            if site_abbreviation not in completion_percentages_per_site.keys():
                completion_percentages_per_site[site_abbreviation] = 0
            cummulative_progress = 0
            for form in item['forms']:
                cummulative_progress += int(form['progress'])
            completion_percentage = cummulative_progress / float(len(item['forms']))
            completion_percentages_per_site[site_abbreviation] += completion_percentage
        return completion_percentages_per_site
        
    def update_nr_records(self, site_abbreviation, nr_records_per_site):
        nr_records = 0
        if site_abbreviation in nr_records_per_site.keys():
            nr_records = nr_records_per_site[site_abbreviation]
        return nr_records
    
    def update_completion_percentage(self, site_, nr_records, completion_rates_per_site):
        completion_rate = 0.0
        if site_ in completion_rates_per_site.keys():
            completion_rate = float(completion_rates_per_site[site_] / float(nr_records)) if nr_records > 0 else 0.0
        return completion_rate
    
    def update_country_code(self, country_id):
        country_codes = self.get_country_codes()
        return country_codes.get_three_digit_code_for(country_id)

    def get_study_sites(self, study_id):
        start_time = current_time_in_seconds()
        study_sites = self.read_from_cache(f'studies/{study_id}/sites')
        if study_sites:
            # LOG.info(f'Core.get_study_sites(): {elapsed_time_in_seconds(start_time)}')
            return study_sites
        with CastorApiClient(self._client_id, self._client_secret, self._token_url, self._api_base_url) as client:

            start_time2 = current_time_in_seconds()
            study_sites_data = client.get_study_sites(study_id)
            LOG.info(f'Core.get_study_sites() client.get_study_sites(): {elapsed_time_in_seconds(start_time2)}')

            start_time2 = current_time_in_seconds()
            participants = client.get_participants(study_id)
            LOG.info(f'Core.get_study_sites() client.get_participants(): {elapsed_time_in_seconds(start_time2)}')

            start_time2 = current_time_in_seconds()
            participant_site_abbreviations = self.get_participant_site_abbreviations(participants)
            LOG.info(f'Core.get_study_sites() client.get_participant_site_abbreviations(): {elapsed_time_in_seconds(start_time2)}')

            start_time2 = current_time_in_seconds()
            participant_progress = client.get_participant_progress(study_id)
            LOG.info(f'Core.get_study_sites() client.get_participant_progress(): {elapsed_time_in_seconds(start_time2)}')

            start_time2 = current_time_in_seconds()
            nr_records_per_site = self.get_nr_records_per_site(participant_progress, participant_site_abbreviations)
            LOG.info(f'Core.get_study_sites() calculate nr. records per site: {elapsed_time_in_seconds(start_time2)}')

            start_time2 = current_time_in_seconds()
            completion_percentages_per_site = self.get_completion_percentages_per_site(participant_progress, participant_site_abbreviations)
            LOG.info(f'Core.get_study_sites() calculate completion percentages per site: {elapsed_time_in_seconds(start_time2)}')

            start_time2 = current_time_in_seconds()
            for study_site_data in study_sites_data:
                study_site_data['nr_records'] = self.update_nr_records(study_site_data['abbreviation'], nr_records_per_site)
                study_site_data['completion_percentage'] = self.update_completion_percentage(study_site_data['abbreviation'], study_site_data['nr_records'], completion_percentages_per_site)
                study_site_data['country_code'] = self.update_country_code(study_site_data['country_id'])
            LOG.info(f'Core.get_study_sites() site info updated: {elapsed_time_in_seconds(start_time2)}')

            study_sites = []
            for study_site_data in study_sites_data:
                study_site = StudySite(study_site_data)
                study_sites.append(study_site)

            self.write_to_cache(f'studies/{study_id}/sites', study_sites)
            LOG.info(f'Core.get_study_sites() elapsed: {elapsed_time_in_seconds(start_time)}')
            return study_sites    
        
    # ASYNCHRONOUS METHODS
        
    def get_countries_async(self, callback):
        self.run_background_task(self.get_countries, callback)

    def get_studies_async(self, callback):
        self.run_background_task(self.get_studies, callback)

    def get_study_async(self, callback, *args):
        self.run_background_task(self.get_study, callback, *args)

    def get_study_sites_async(self, callback, *args):
        self.run_background_task(self.get_study_sites, callback, *args)

    # HELPERS

    def run_background_task(self, func, callback, *args, **kwargs):
        future = self._executor.submit(func, *args, **kwargs)
        with self._callbacks_lock:
            self._callbacks[future] = callback
        future.add_done_callback(self.handle_task_done)

    def handle_task_done(self, future):
        result, error = None, None
        try:
            result = future.result()
        except Exception as e:
            error = e
        with self._callbacks_lock:
            callback = self._callbacks.pop(future, None)
        if callback:
            callback(result, error)


import re


class Route:
    def __init__(self, page, path_pattern):
        self._page = page
        self._path_pattern = path_pattern
        self._param_names = []
        self._regex = self.compile_pattern(path_pattern)

    def get_page(self):
        return self._page

    def compile_pattern(self, path_pattern):
        pattern = re.sub(r":(\w+)", r"(?P<\1>[^/]+)", path_pattern)
        self._param_names = re.findall(r":(\w+)", path_pattern)
        return re.compile(f"^{pattern}$")

    def match(self, path):
        match = self._regex.match(path)
        if match:
            return match.groupdict()
        return None


from sqlalchemy import create_engine

from castoranalytics.core.singleton import singleton
from castoranalytics.core.db.models.basemodel import BaseModel


@singleton
class Engine:
    def __init__(self):
        self._engine = create_engine(f'sqlite:///db.sqlite3', echo=False)
        BaseModel.metadata.create_all(self._engine)

    def get(self):
        return self._engine


import os
import requests

from PySide6.QtWidgets import (
    QMainWindow,
    QStackedLayout,
    QWidget,
    QVBoxLayout,
)
from PySide6.QtGui import (
    QGuiApplication,
    QIcon,
    QAction,
)
from PySide6.QtCore import Qt

import castoranalytics.ui.constants as constants

from castoranalytics.ui.pages.router import Router
from castoranalytics.ui.pages.studylistpage import StudyListPage
from castoranalytics.ui.pages.studypage import StudyPage
from castoranalytics.ui.pages.studysitelistpage import StudySiteListPage
from castoranalytics.ui.pages.settingspage import SettingsPage
from castoranalytics.ui.utils import Label, resource_path
from castoranalytics.ui.components.backgroundimage import BackgroundImage
from castoranalytics.core.logging import LogManager

LOG = LogManager()


class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self._version = self.init_version()
        self._app_label_text = constants.CASTOR_ANALYTICS_WINDOW_TITLE + f' {self._version}'
        self._background_image = self.init_background_image()
        self._router = self.init_router()
        self._file_menu = self.init_file_menu()
        self._page_layout = self.init_page_layout(self._app_label_text, self._background_image, self._router)
        self.init_main_window(self._app_label_text, self._page_layout)
        self.show()

    # INITIALIZATION

    def init_version(self):
        with open(resource_path(os.path.join(constants.CASTOR_ANALYTICS_RESOURCES_DIR, 'VERSION')), 'r') as f:
            version = f.readline().strip()
            return version
        
    def init_background_image(self):
        image = BackgroundImage()
        return image

    def init_router(self):
        router = Router()
        router.add_page(StudyListPage(), '/studies')
        router.add_page(StudyPage(), '/studies/:study_id')
        router.add_page(StudySiteListPage(), '/studies/:study_id/sites')
        router.add_page(SettingsPage(), '/settings')
        # router.navigate('/studies')
        return router

    def init_file_menu(self):
        menu = self.menuBar().addMenu('File')
        menu_settings_action = QAction('Settings', self)
        menu_settings_action.triggered.connect(self.on_open_settings_page)
        menu_exit_action = QAction('Exit', self)
        menu_exit_action.triggered.connect(self.close)
        menu.addAction(menu_settings_action)
        menu.addAction(menu_exit_action)
        return menu

    def init_page_layout(self, app_label_text, background_image, router):
        app_label = Label(app_label_text, type=Label.HEADING1)
        app_label.setAlignment(Qt.AlignCenter)
        page_widget = QWidget()
        layout = QStackedLayout(page_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setStackingMode(QStackedLayout.StackAll)
        layout.addWidget(background_image)
        layout.addWidget(router)
        page_layout = QVBoxLayout()
        page_layout.addWidget(app_label)
        page_layout.addWidget(page_widget)
        return page_layout

    def init_main_window(self, app_label_text, page_layout):
        widget = QWidget()
        widget.setLayout(page_layout)
        self.setWindowTitle(app_label_text)
        self.setWindowIcon(QIcon(resource_path(os.path.join(
            constants.CASTOR_ANALYTICS_RESOURCES_IMAGES_ICONS_DIR, constants.CASTOR_ANALYTICS_RESOURCES_ICON))))
        self.setCentralWidget(widget)
        self.resize(constants.CASTOR_ANALYTICS_WINDOW_W, constants.CASTOR_ANALYTICS_WINDOW_H)
        self.center_window()

    # EVENT HANDLERS

    def resizeEvent(self, event):
        self._background_image.rescale()
        self._router.navigate('/studies')
        return super().resizeEvent(event)

    def on_open_settings_page(self):
        self._router.navigate('/settings')

    # MISCELLANEOUS

    def center_window(self):
        screen = QGuiApplication.primaryScreen().geometry()
        x = (screen.width() - self.geometry().width()) / 2
        y = (screen.height() - self.geometry().height()) / 2
        self.move(int(x), int(y))


from PySide6.QtWidgets import (
    QPushButton,
    QFormLayout,
    QLineEdit,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
)

import castoranalytics.ui.constants as constants

from castoranalytics.ui.pages.basepage import BasePage
from castoranalytics.ui.utils import Label


class SettingsPage(BasePage):
    def __init__(self):
        super(SettingsPage, self).__init__(name='Settings')
        self._back_button = self.init_back_button()
        self._save_button = self.init_save_button()
        self._client_id_field = self.init_client_id_field()
        self._client_secret_field = self.init_client_secret_field()
        self._token_url_field = self.init_token_url_field()
        self._api_base_url_field = self.init_api_base_url_field()
        self._api_settings_form = self.init_api_settings_form(
            self._client_id_field, self._client_secret_field, self._token_url_field, self._api_base_url_field
        )
        self._license_key_field = self.init_license_key_field()
        self._license_key_form = self.init_license_key_form(self._license_key_field)
        self.init_page_layout()

    # INITIALIZATION

    def init_back_button(self):
        button = QPushButton('Back', self)
        button.clicked.connect(self.on_back)
        return button
    
    def init_save_button(self):
        button = QPushButton('Save settings', self)
        button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
        button.clicked.connect(self.on_save)
        return button
    
    def init_client_id_field(self):
        return QLineEdit(self.get_setting(constants.CASTOR_ANALYTICS_SETTINGS_KEY_CLIENT_ID))
    
    def init_client_secret_field(self):
        return QLineEdit(self.get_setting(constants.CASTOR_ANALYTICS_SETTINGS_KEY_CLIENT_SECRET))
    
    def init_token_url_field(self):
        return QLineEdit(self.get_setting(
            constants.CASTOR_ANALYTICS_SETTINGS_KEY_TOKEN_URL, default=constants.CASTOR_ANALYTICS_SETTINGS_KEY_TOKEN_URL_DEFAULT))
    
    def init_api_base_url_field(self):
        return QLineEdit(self.get_setting(
            constants.CASTOR_ANALYTICS_SETTINGS_KEY_API_BASE_URL, default=constants.CASTOR_ANALYTICS_SETTINGS_KEY_API_BASE_URL_DEFAULT))
    
    def init_license_key_field(self):
        return QLineEdit(self.get_setting(
            constants.CASTOR_ANALYTICS_SETTINGS_KEY_LICENSE_KEY))
    
    def init_api_settings_form(self, client_id_field, client_secret_field, token_url_field, api_base_url_field):
        form = QFormLayout()
        form.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow) # needed on MacOS
        form.addRow(constants.CASTOR_ANALYTICS_SETTINGS_KEY_CLIENT_ID_NAME, client_id_field)
        form.addRow(constants.CASTOR_ANALYTICS_SETTINGS_KEY_CLIENT_SECRET_NAME, client_secret_field)
        form.addRow(constants.CASTOR_ANALYTICS_SETTINGS_KEY_TOKEN_URL_NAME, token_url_field)
        form.addRow(constants.CASTOR_ANALYTICS_SETTINGS_KEY_API_BASE_URL_NAME, api_base_url_field)
        layout = QVBoxLayout()
        layout.addWidget(Label(constants.CASTOR_ANALYTICS_API_SETTINGS_TITLE, type=Label.HEADING1))
        layout.addLayout(form)
        return layout
    
    def init_license_key_form(self, license_key_field):
        form = QFormLayout()
        form.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow) # needed on MacOS
        form.addRow(constants.CASTOR_ANALYTICS_SETTINGS_KEY_LICENSE_KEY_NAME, license_key_field)
        layout = QVBoxLayout()
        layout.addWidget(Label(constants.CASTOR_ANALYTICS_LICENSE_KEY_SETTINGS_TITLE, type=Label.HEADING1))
        layout.addLayout(form)
        return layout
    
    def init_page_layout(self):
        self.get_layout().addWidget(self._back_button)
        self.get_layout().addLayout(self._api_settings_form)
        # self.get_layout().addLayout(self._license_key_form)
        self.get_layout().addItem(QSpacerItem(0, 20, QSizePolicy.Minimum, QSizePolicy.Expanding))
        self.get_layout().addWidget(self._save_button)

    # EVENT HANDLERS

    def on_back(self):
        self.back()

    def on_save(self):
        client_id = self._client_id_field.text()
        client_secret = self._client_secret_field.text()
        token_url = self._token_url_field.text()
        api_base_url = self._api_base_url_field.text()
        self._settings.setValue(constants.CASTOR_ANALYTICS_SETTINGS_KEY_CLIENT_ID, client_id)
        self._settings.setValue(constants.CASTOR_ANALYTICS_SETTINGS_KEY_CLIENT_SECRET, client_secret)
        self._settings.setValue(constants.CASTOR_ANALYTICS_SETTINGS_KEY_TOKEN_URL, token_url)
        self._settings.setValue(constants.CASTOR_ANALYTICS_SETTINGS_KEY_API_BASE_URL, api_base_url)
        self.get_core().set_api_credentials(client_id, client_secret, token_url, api_base_url)
        self.back()

    def on_navigate(self, params):
        pass


from castoranalytics.ui.utils import is_macos


# Window settings
CASTOR_ANALYTICS_WINDOW_TITLE = 'Castor Analytics'
CASTOR_ANALYTICS_WINDOW_W = 1024
CASTOR_ANALYTICS_WINDOW_H = 600

# Resources
CASTOR_ANALYTICS_RESOURCES_DIR = 'castoranalytics/ui/resources'
CASTOR_ANALYTICS_RESOURCES_IMAGES_DIR = 'castoranalytics/ui/resources/images'
CASTOR_ANALYTICS_RESOURCES_IMAGES_ICONS_DIR = 'castoranalytics/resources/ui/images/icons'
CASTOR_ANALYTICS_RESOURCES_ICON = 'castoranalytics.icns' if is_macos() else 'castoranalytics.ico'
CASTOR_ANALYTICS_RESOURCES_BACKGROUND_IMAGE = 'home.png'
CASTOR_ANALYTICS_RESOURCES_BACKGROUND_IMAGE_OPACITY = 0.25

# API Settings
CASTOR_ANALYTICS_API_SETTINGS_TITLE = 'API settings'
CASTOR_ANALYTICS_API_SETTINGS_ERROR_MESSAGE = 'It looks like your API settings are incomplete. Please go to settings.'
CASTOR_ANALYTICS_SETTINGS_KEY_CLIENT_ID = 'castoranalytics.client_id'
CASTOR_ANALYTICS_SETTINGS_KEY_CLIENT_ID_NAME = 'API client ID'
CASTOR_ANALYTICS_SETTINGS_KEY_CLIENT_SECRET = 'castoranalytics.client_secret'
CASTOR_ANALYTICS_SETTINGS_KEY_CLIENT_SECRET_NAME = 'API client secret'
CASTOR_ANALYTICS_SETTINGS_KEY_TOKEN_URL = 'castoranalytics.token_url'
CASTOR_ANALYTICS_SETTINGS_KEY_TOKEN_URL_NAME = 'API token URL'
CASTOR_ANALYTICS_SETTINGS_KEY_TOKEN_URL_DEFAULT = 'https://data.castoredc.com/oauth/token'
CASTOR_ANALYTICS_SETTINGS_KEY_API_BASE_URL = 'castoranalytics.api_base_url'
CASTOR_ANALYTICS_SETTINGS_KEY_API_BASE_URL_NAME = 'API base URL'
CASTOR_ANALYTICS_SETTINGS_KEY_API_BASE_URL_DEFAULT = 'https://data.castoredc.com/api'

# License key settings
CASTOR_ANALYTICS_LICENSE_KEY_SETTINGS_TITLE = 'License key'
CASTOR_ANALYTICS_LICENSE_KEY_SETTINGS_ERROR_MESSAGE = 'It looks like you have not set the license key. Please go to settings.'
CASTOR_ANALYTICS_SETTINGS_KEY_LICENSE_KEY = 'castoranalytics.license_key'
CASTOR_ANALYTICS_SETTINGS_KEY_LICENSE_KEY_NAME = 'License key'

# Studies and sites
CASTOR_ANALYTICS_STUDY_SITES_WARNING_MAX = 10
CASTOR_ANALYTICS_STUDY_SITES_WARNING = f'Your study has >{CASTOR_ANALYTICS_STUDY_SITES_WARNING_MAX} sites. Loading these for the first time will take a few minutes'


from PySide6.QtWidgets import (
    QTableWidget,
    QTableWidgetItem,
    QHeaderView,
    QSizePolicy,
)
from PySide6.QtCore import Qt

from castoranalytics.ui.pages.basepage import BasePage
from castoranalytics.ui.utils import Label
from castoranalytics.core.logging import LogManager

LOG = LogManager()


class StudyListPage(BasePage):
    def __init__(self):
        super(StudyListPage, self).__init__(name='Studies')
        self._study_list_label = self.init_study_list_label()
        self._table_widget = self.init_table_widget()
        self.init_page_layout()

    # INITIALIZATION

    def init_study_list_label(self):
        label = Label('Studies', type=Label.HEADING1)
        return label

    def init_table_widget(self):
        widget = QTableWidget()
        widget.setSortingEnabled(True)
        widget.horizontalHeader().setVisible(False)
        widget.verticalHeader().setVisible(False)
        widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        widget.itemClicked.connect(self.on_study_selected)
        return widget

    def init_page_layout(self):
        self.get_layout().addWidget(Label('Studies', type=Label.HEADING1))
        self.get_layout().addWidget(self._table_widget)

    # TABLE

    def update_study_list_label(self, studies, error):
        if error:
            self._study_list_label.setText(error)
        else:
            self._study_list_label.setText('Studies')

    def update_table_widget(self, studies):
        self._table_widget.setRowCount(len(studies))
        self._table_widget.setColumnCount(1)
        for row, study in enumerate(studies):
            item = QTableWidgetItem(study.get_name())
            item.setData(Qt.UserRole, study)
            self._table_widget.setItem(row, 0, item)
        self._table_widget.setHorizontalHeaderLabels(['Study name'])
        self._table_widget.setAlternatingRowColors(True)
        self._table_widget.sortItems(0, Qt.AscendingOrder)
        self._table_widget.horizontalHeader().setSectionResizeMode(0, QHeaderView.Stretch)
        self._table_widget.horizontalHeader().setDefaultAlignment(Qt.AlignLeft | Qt.AlignVCenter)

    # EVENT HANDLERS

    def on_study_selected(self, item):
        self.navigate(f'/studies/{item.data(Qt.UserRole).get_id()}')

    def on_data_ready(self, studies, error):
        if error:
            self._study_list_label.setText(error)
        self.update_table_widget(studies)

    def on_navigate(self, params):
        self._table_widget.clearContents()
        self.load_data('get_studies')


import pandas as pd

from PySide6.QtWidgets import (
    QTableWidget,
    QTableWidgetItem,
    QHeaderView,
    QSizePolicy,
    QPushButton,
    QFileDialog,
    QMessageBox,
)
from PySide6.QtCore import Qt

from castoranalytics.ui.pages.basepage import BasePage
from castoranalytics.ui.components.numerictablewidgetitem import NumericTableWidgetItem
from castoranalytics.ui.utils import Label
from castoranalytics.core.logging import LogManager

LOG = LogManager()


class StudySiteListPage(BasePage):
    def __init__(self):
        super(StudySiteListPage, self).__init__(name='Study sites')
        self._study_id = None
        self._study = None
        self._study_sites = None
        self._back_button = self.init_back_button()
        self._export_to_excel_button = self.init_export_to_excel_button()
        self._study_site_list_label = self.init_study_site_list_label()
        self._table_widget = self.init_table_widget()
        self.init_page_layout()

    # INITIALIZATION

    def init_back_button(self):
        button = QPushButton('Back', self)
        button.clicked.connect(self.on_back)
        return button
    
    def init_export_to_excel_button(self):
        button = QPushButton('Export site data to Excel', self)
        button.clicked.connect(self.on_export_to_excel)
        return button

    def init_study_site_list_label(self):
        label = Label('', type=Label.HEADING1)
        return label
    
    def init_table_widget(self):
        widget = QTableWidget()
        widget.setSortingEnabled(True)
        widget.verticalHeader().setVisible(False)
        widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        widget.itemClicked.connect(self.on_study_site_selected)
        return widget

    def init_page_layout(self):
        self.get_layout().addWidget(self._back_button)
        self.get_layout().addWidget(self._export_to_excel_button)
        self.get_layout().addWidget(self._study_site_list_label)
        self.get_layout().addWidget(self._table_widget)

    # TABLE

    def update_study_site_list_label(self, error):
        if error:
            self._study_site_list_label.setText(error)
        else:
            self._study_site_list_label.setText(f'Study sites for {self._study.get_name()}')

    # def update_table_widget(self, study_sites):
    #     self._table_widget.clearContents()
    #     self._table_widget.setRowCount(len(study_sites))
    #     self._table_widget.setColumnCount(5)
    #     self._table_widget.setHorizontalHeaderLabels(['Site code', 'Site name', 'Country code', 'Nr. records', 'Complete %'])
    #     self._table_widget.setAlternatingRowColors(True)
    #     self._table_widget.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
    #     self._table_widget.horizontalHeader().setDefaultAlignment(Qt.AlignLeft | Qt.AlignVCenter)
    #     for row, study_site in enumerate(study_sites):
    #         self._table_widget.setItem(row, 0, QTableWidgetItem(study_site.get_abbreviation()))
    #         self._table_widget.setItem(row, 1, QTableWidgetItem(study_site.get_name()))
    #         self._table_widget.setItem(row, 2, QTableWidgetItem(study_site.get_country_code()))
    #         self._table_widget.setItem(row, 3, NumericTableWidgetItem(study_site.get_nr_records())) # Handles sorting correctly for numbers
    #         self._table_widget.setItem(row, 4, NumericTableWidgetItem(study_site.get_completion_percentage()))
        
    def update_table_widget(self, study_sites):
        self._table_widget.setSortingEnabled(False)
        self._table_widget.clear()
        self._table_widget.setRowCount(len(study_sites))
        self._table_widget.setColumnCount(5)
        self._table_widget.setHorizontalHeaderLabels(['Site code', 'Site name', 'Country code', 'Nr. records', 'Complete %'])
        self._table_widget.setAlternatingRowColors(True)
        self._table_widget.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self._table_widget.horizontalHeader().setDefaultAlignment(Qt.AlignLeft | Qt.AlignVCenter)
        for idx, study_site in enumerate(study_sites):
            self._table_widget.setItem(idx, 0, QTableWidgetItem(study_site.get_abbreviation()))
            self._table_widget.setItem(idx, 1, QTableWidgetItem(study_site.get_name()))
            self._table_widget.setItem(idx, 2, QTableWidgetItem(study_site.get_country_code()))
            self._table_widget.setItem(idx, 3, NumericTableWidgetItem(study_site.get_nr_records()))
            self._table_widget.setItem(idx, 4, NumericTableWidgetItem(study_site.get_completion_percentage()))    
        self._table_widget.setSortingEnabled(True)
        self._table_widget.viewport().update()
    
    # EVENT HANDLERS

    def on_back(self):
        self.navigate(f'/studies/{self._study_id}')

    def on_export_to_excel(self):
        if self._study_sites:
            file_path, _ = QFileDialog.getSaveFileName(self, 'Save file as...', '.', filter='Excel (.xlsx)')
            if file_path:
                if not file_path.endswith('.xlsx'):
                    QMessageBox.warning(self, 'Error', 'File must have extension .xlsx')
                    return
                data = {
                    'name': [],
                    'abbreviation': [],
                    'country_code': [],
                    'nr_records': [],
                    'completion_percentage': [],
                }
                for idx, study_site in enumerate(self._study_sites):
                    data['name'].append(study_site.get_name())
                    data['abbreviation'].append(study_site.get_abbreviation())
                    data['country_code'].append(study_site.get_country_code())
                    data['nr_records'].append(study_site.get_nr_records())
                    data['completion_percentage'].append(study_site.get_completion_percentage())
                df = pd.DataFrame(data)
                df.to_excel(file_path, index=False, engine="openpyxl")
                QMessageBox.information(self, 'Info', 'Site data successfully saved')

    def on_study_site_selected(self, item):
        # self.navigate(f'/studies/{item.data(Qt.UserRole).get_id()}')
        pass

    def on_navigate(self, params):
        self._study_site_list_label.setText('')
        self._table_widget.clearContents()
        self._study_id = params.get('study_id', None)
        if self._study_id:
            self._study = self.get_core().get_study(self._study_id)
            self.load_data('get_study_sites', self._study_id)
    
    def on_data_ready(self, study_sites, error):
        self._study_sites = study_sites
        self.update_study_site_list_label(error)
        self.update_table_widget(self._study_sites)


import os

from PySide6.QtWidgets import (
    QLabel,
    QSizePolicy
)
from PySide6.QtGui import (
    QPixmap,
    QPainter, 
    QColor,
)
from PySide6.QtCore import Qt

import castoranalytics.ui.constants as constants
from castoranalytics.ui.utils import resource_path


class BackgroundImage(QLabel):
    def __init__(self):
        super(BackgroundImage, self).__init__()
        self.setAlignment(Qt.AlignCenter)
        self.setScaledContents(True)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        image_path = resource_path(os.path.join(
            constants.CASTOR_ANALYTICS_RESOURCES_IMAGES_DIR, constants.CASTOR_ANALYTICS_RESOURCES_BACKGROUND_IMAGE))
        self.setPixmap(self.apply_opacity_to_pixmap(QPixmap(image_path), constants.CASTOR_ANALYTICS_RESOURCES_BACKGROUND_IMAGE_OPACITY))
        self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents, True) # should not handle events!
        self._background_image_pixmap = None

    def rescale(self):
        if self._background_image_pixmap is not None:
            scaled_pixmap = self._background_image_pixmap.scaled(
                self.size(), aspectMode=Qt.AspectRatioMode.IgnoreAspectRatio, mode=Qt.TransformationMode.SmoothTransformation)
            self.setPixmap(
                self.apply_opacity_to_pixmap(
                    scaled_pixmap, constants.CASTOR_ANALYTICS_RESOURCES_BACKGROUND_IMAGE_OPACITY))

    def apply_opacity_to_pixmap(self, pixmap, opacity):
        transparent_pixmap = QPixmap(pixmap.size())
        transparent_pixmap.fill(QColor(0, 0, 0, 0))
        painter = QPainter(transparent_pixmap)
        painter.setOpacity(opacity)
        painter.drawPixmap(0, 0, pixmap)
        painter.end()
        return transparent_pixmap


import sys
import datetime


class LogManager:
    def __init__(self):
        self._name = 'castormobile'

    def _log(self, level, message):
        timestamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        message = f'[{timestamp}] {level} {self._name}: {message}'
        if hasattr(sys, '_MEIPASS'):
            with open('CastorAnalytics.log', 'w', encoding='utf-8') as f:
                f.write(message + '\n')
        print(message)

    def info(self, message):
        self._log('INFO', message)

    def warning(self, message):
        self._log('WARNING', message)

    def error(self, message):
        self._log('ERROR', message)


class StudySite:
    def __init__(self, study_site_data):
        self._study_site_data = study_site_data

    def get_id(self):
        return self._study_site_data['site_id']
    
    def get_name(self):
        return self._study_site_data['name']

    def get_abbreviation(self):
        return self._study_site_data['abbreviation']

    def get_country_id(self):
        return self._study_site_data['country_id']
    
    def get_country_code(self):
        return self._study_site_data['country_code']

    def get_nr_records(self):
        return self._study_site_data['nr_records']

    def get_completion_percentage(self):
        return int(round(self._study_site_data['completion_percentage'], 0))
    
    def __str__(self):
        return 'StudySite(name={}, abbreviation={}, country_id={}, country_code={}, nr_records={}, completion_percentage={})'.format(
            self.get_name(),
            self.get_abbreviation(),
            self.get_country_id(),
            self.get_country_code(), 
            self.get_nr_records(),
            self.get_completion_percentage(),
        )


import time
import math


def current_time_in_milliseconds():
    return int(round(time.time() * 1000))


def current_time_in_seconds():
    return int(round(current_time_in_milliseconds() / 1000.0))


def elapsed_time_in_milliseconds(start_time_in_milliseconds):
    return current_time_in_milliseconds() - start_time_in_milliseconds


def elapsed_time_in_seconds(start_time_in_seconds):
    return current_time_in_seconds() - start_time_in_seconds


def duration(seconds):
    h = int(math.floor(seconds/3600.0))
    remainder = seconds - h * 3600
    m = int(math.floor(remainder/60.0))
    remainder = remainder - m * 60
    s = int(math.floor(remainder))
    return '{} hours, {} minutes, {} seconds'.format(h, m, s)


from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QLabel,
)
from PySide6.QtCore import Qt

import castoranalytics.ui.constants as constants

from castoranalytics.core import Core
from castoranalytics.core.logging import LogManager
from castoranalytics.ui.utils import to_main_thread, BusyOverlay, Settings

LOG = LogManager()


class BasePage(QWidget):
    def __init__(self, name):
        super(BasePage, self).__init__()
        self._name = name
        self._settings = Settings()
        self._busy_overlay = BusyOverlay(self)
        self._router = None
        self._core = self.init_core()
        self._error_label = self.init_error_label()
        self._page_layout = self.init_page_layout_internal(self._error_label)
        self.setLayout(self._page_layout)

    def init_error_label(self):
        label = QLabel()
        if not self.get_core().ready():
            label.setText(constants.CASTOR_ANALYTICS_API_SETTINGS_ERROR_MESSAGE)
        return label
    
    def init_page_layout_internal(self, error_label):
        layout = QVBoxLayout(self)
        layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        layout.addWidget(error_label)
        return layout

    def init_core(self):
        self._core = Core(
            self.get_setting(constants.CASTOR_ANALYTICS_SETTINGS_KEY_CLIENT_ID, None),
            self.get_setting(constants.CASTOR_ANALYTICS_SETTINGS_KEY_CLIENT_SECRET, None),
            self.get_setting(constants.CASTOR_ANALYTICS_SETTINGS_KEY_TOKEN_URL, None),
            self.get_setting(constants.CASTOR_ANALYTICS_SETTINGS_KEY_API_BASE_URL, None),
        )
        return self._core
    
    def get_core(self):
        return self._core

    def get_layout(self):
        return self._page_layout
    
    def load_data(self, func_name, *args, **kwargs):
        func_name = func_name if func_name.endswith('_async') else func_name + '_async'
        try:
            if self.get_core().ready():
                func = getattr(self.get_core(), func_name)
                self._busy_overlay.show_overlay()
                func(to_main_thread(self.on_data_ready_internal), *args, **kwargs)
                self._error_label.setText('')
            else:
                self._error_label.setText(constants.CASTOR_ANALYTICS_API_SETTINGS_ERROR_MESSAGE)
        except Exception as e:
            self._busy_overlay.hide_overlay()
            LOG.error(e)

    def on_data_ready_internal(self, result, error):
        self.on_data_ready(result, error)
        self._busy_overlay.hide_overlay()

    def on_data_ready(self, result, error):
        raise NotImplementedError()

    def set_router(self, router):
        self._router = router

    def back(self):
        if self._router:
            self._router.back()

    def navigate(self, path):
        if self._router:
            self._router.navigate(path)

    def get_setting(self, name, default=None):
        return self._settings.get(name, default)
    
    def set_setting(self, name, value):
        self._settings.set(name, value)

    def on_navigate(self, params):
        raise NotImplementedError()


class Study:
    def __init__(self, study_data):
        self._study_data = study_data

    def get_id(self):
        return self._study_data['study_id']

    def get_name(self):
        return self._study_data['name']
    
    def get_created_on(self):
        return self._study_data['created_on']
    
    def __str__(self):
        return 'Study(id={}, name={}, created_on={})'.format(
            self.get_id(),
            self.get_name(),
            self.get_created_on(),
        )


class Country:
    def __init__(self, country_data):
        self._country_data = country_data

    def get_id(self):
        return int(self._country_data['country_id'])
    
    def get_name(self):
        return self._country_data['country_name']
    
    def get_tld_code(self):
        return self._country_data['country_tld']

    def get_two_digit_code(self):
        return self._country_data['country_cca2']
    
    def get_three_digit_code(self):
        return self._country_data['country_cca3']